; convers.inc                                                              2012-02-01 AgF
;
; Define universal test code to test latency and throughput for conversion instructions with
; different register types for input and output
; (c) Copyright 2012 by Agner Fog. GNU General Public License www.gnu.org/licenses
;
; Parameters:
;
; instruct1:   Instruction to convert from A to B
;
; instruct2:   Instruction to convert from B to A
;
; instruct3:   Possible extra chain instruction with A type registers for both input and output
;
; regtype1:    Register type for operand type A (r8, r16, r32, r64, mmx, xmm, ymm, m, m8, m16, m32, m64, m128, m256)
;
; regtype2:    Register type for operand type B (r8, r16, r32, r64, mmx, xmm, ymm, m, m8, m16, m32, m64, m128, m256)
;
; numimm:      Number of immediate operands for instruct1 (0 - 1). Default = 0
;
; immvalue:    Value of first immediate operand for instruct1. Default = 0
;
; tmode:       Test mode:
;              T1:   measure throughput of instruct1 B, A
;
;              L1:   measure latency of instruct1 A, A if regtype1 and regtype2 are the same
;
;              L12:  measure latency of instruct1 B, A + instruct2 A, B
;
;              L123: measure latency of instruct1 B, A + instruct2 A, B + instruct3 A, A
;
;              LSX:  measure latency of movsx or movzx with memory operand
;
; -------------------------------------------------------------------------------------------------

; select register names of desired types:

%ifidni regtype1, r8
   %define RA1 al
   %define RA2 bl
%elifidni regtype1, r8h
   %define RA1 ah
   %define RA2 bh
%elifidni regtype1, r16
   %define RA1 ax
   %define RA2 bx
%elifidni regtype1, r32
   %define RA1 eax
   %define RA2 ebx
%elifidni regtype1, r64
   %define RA1 rax
   %define RA2 rbx
%elifidni regtype1, mmx
   %define RA1 mm0
   %define RA2 mm1
%elifidni regtype1, xmm
   %define RA1 xmm0
   %define RA2 xmm1
%elifidni regtype1, ymm
   %define RA1 ymm0
   %define RA2 ymm1
%elifidni regtype1, m
   %define RA1 [rsi]
   %define RA2 [rdi]
%elifidni regtype1, m8
   %define RA1 byte [rsi]
   %define RA2 byte [rdi]
%elifidni regtype1, m16
   %define RA1 word [rsi]
   %define RA2 word [rdi]
%elifidni regtype1, m32
   %define RA1 dword [rsi]
   %define RA2 dword [rdi]
%elifidni regtype1, m64
   %define RA1 qword [rsi]
   %define RA2 qword [rdi]
%elifidni regtype1, m128
   %define RA1 oword [rsi]
   %define RA2 oword [rdi]
%elifidni regtype1, m256
   %define RA1 yword [rsi]
   %define RA2 yword [rdi]
%else
   %error unknown regtype1
%endif

%ifndef regtype2
   %define regtype2 regtype1
%endif

%ifidni regtype2, r8
   %define RB1 al
   %define RB2 bl
%elifidni regtype2, r8h
   %define RB1 ah
   %define RB2 bh
%elifidni regtype2, r16
   %define RB1 ax
   %define RB2 bx
%elifidni regtype2, r32
   %define RB1 eax
   %define RB2 ebx
%elifidni regtype2, r64
   %define RB1 rax
   %define RB2 rbx
%elifidni regtype2, mmx
   %define RB1 mm0
   %define RB2 mm1
%elifidni regtype2, xmm
   %define RB1 xmm0
   %define RB2 xmm1
%elifidni regtype2, ymm
   %define RB1 ymm0
   %define RB2 ymm1
%elifidni regtype2, m
   %define RB1 [rsi]
   %define RB2 [rdi]
%elifidni regtype2, m8
   %define RB1 byte [rsi]
   %define RB2 byte [rdi]
%elifidni regtype2, m16
   %define RB1 word [rsi]
   %define RB2 word [rdi]
%elifidni regtype2, m32
   %define RB1 dword [rsi]
   %define RB2 dword [rdi]
%elifidni regtype2, m64
   %define RB1 qword [rsi]
   %define RB2 qword [rdi]
%elifidni regtype2, m128
   %define RB1 oword [rsi]
   %define RB2 oword [rdi]
%elifidni regtype2, m256
   %define RB1 yword [rsi]
   %define RB2 yword [rdi]
%else
   %error unknown regtype1
%endif

; initialization of registers
%macro testinit2 0
   xor eax, eax
%endmacro


; test code for each mode:

%ifidni tmode,T1        ; measure throughput of instruct1 B, A   

   %macro testcode 0
      instruct1 RB1, RA2 immoperands1
   %endmacro

%elifidni tmode,L1        ; measure latency of instruct1 A, A

   %macro testcode 0
      instruct1 RB1, RA1 immoperands1
   %endmacro

%elifidni tmode,L12        ; measure latency of instruct1 B, A + instruct2 A, B

   %macro testcode 0
      instruct1 RB2, RA1 immoperands1
      instruct2 RA1, RB2
   %endmacro

%elifidni tmode,L123        ; measure latency of instruct1 B, A + instruct2 A, B + instruct3 A, A

   %macro testcode 0
      instruct1 RB2, RA1 immoperands1
      instruct2 RA1, RB2
      instruct3 RA1, RA1
   %endmacro

%elifidni tmode,LSX         ; measure latency of movsx r,[m] + mov [m],r

   %macro testcode 0
      instruct1 RB1, RA1
      %ifidni regtype1,m8
         mov byte [rsi], al
      %elifidni regtype1,m16
         mov word [rsi], ax
      %elifidni regtype1,m32
         mov dword [rsi], eax
      %else
         %error unknown combination of operands
      %endif
   %endmacro

%else
   %error unknown tmode
%endif

